
import { Service } from "../service"
import * as mongodb from "mongodb"
import * as Leaf from "leaf-ts"
import * as express from "express"
import * as crypto from "crypto"
import * as fs from "fs"
import * as pathModule from "path"

export type Callback<T = any, E extends Error = Error> = (e?: E, result?: T) => void
export const Errors = Leaf.ErrorDoc.build({
    blobError: {},
    unknowError: {},
    notFound: {}
})

export interface Blob {
    name: string
    hash: string
    mime?: string
    format?: string
    id: string   // auto incremental
    size?: number
    url?: string
}

export class BlobService extends Service {
    readonly name = "BlobService"
    public blobCollection: mongodb.Collection
    readonly dependencies = ["IncrementalIdService", "MongodbService", "ExpressService"]
    readonly collectionName = "blob"
    services: { MongodbService, IncrementalIdService, ExpressService }
    route: string = ""
    path: string = ""

    constructor(private readonly settings: {
        blob: {
            webAccessRoute: string
            blobStorageDir: string
        }
    }) {
        super()
        this.configRouteAndPath()
    }

    private configRouteAndPath() {
        if (this.settings.blob && this.settings.blob.blobStorageDir) {
            this.path = this.settings.blob.blobStorageDir
        } else {
            this.path = pathModule.join(__dirname, "../../../../../blob")
        }

        if (this.settings.blob && this.settings.blob.webAccessRoute) {
            this.route = this.settings.blob.webAccessRoute
        } else {
            this.route = "/blob:id"
        }
    }


    initialize(done) {
        this.blobCollection = this.services.MongodbService.db.collection(this.collectionName)
        this.services.IncrementalIdService.ensure({
            name: "blob",
            offset: 100000
        }, () => {
            fs.stat(this.path, (err) => {
                if (err) {
                    fs.mkdirSync(this.path)
                }
            })
            done()
        })

        this.services.ExpressService.server.use(this.route, (req, res) => {
            let id = req.params.id
            this.getBlob({ id: id }, (err, blob) => {
                if (err || blob == null) {
                    res.status(404).send(err)
                } else {
                    let path = `${this.path}/${blob.hash}.${blob.format}`
                    if (req.query.download) {
                        let stats = fs.statSync(path);
                        if (stats.isFile()) {
                            res.set({
                                'Content-Type': 'application/octet-stream',
                                'Content-Disposition': 'attachment; filename=' + encodeURIComponent(blob.name),
                                'Content-Length': stats.size
                            });
                            fs.createReadStream(path).pipe(res);
                        } else {
                            res.status(404).send("download resource is not exist")
                        }
                    } else {
                        fs.stat(path, (err, result) => {
                            if (err) {
                                res.status(404)
                                res.send(err)
                            } else {
                                res.status(200).sendFile(path)
                            }
                        })
                    }

                }
            })
        })
    }

    redirectToServerPath(req: express.Request, res: express.Response) {
        let id = req.params.blobId
        //console.log("redirectId", id)
        this.getBlob(id, (err, result) => {
            if (err) {
                res.end()
            } else {
                res.redirect(result.url)
            }
        })
    }


    createBlob(option: { path: string, name: string }, callback: Callback<Blob>) {
        this.createBlobFileByHardLink(option.path, (err, blob: Blob) => {
            if (err) {
                callback(err)
                return
            }
            this.services.IncrementalIdService.next({ name: "blob" }, (err, id) => {
                if (err) {
                    callback(new Errors.unknowError("Fail to create id", { via: err }))
                    return
                }
                blob.name = option.name
                blob.id = id.toString()
                blob.url = `/blob/${blob.id}`
                this.blobCollection.insert(blob, (err, result) => {
                    if (err) {
                        callback(new Errors.unknowError("Fail to insert blob", { via: err }))
                        return
                    }
                    callback(null, blob)
                })
            })
        })
    }

    private createBlobFileByHardLink(path, callback: Callback<Blob>) {
        let hash = crypto.createHash("md5")
        let stream = fs.createReadStream(path)
        let blob = {} as Blob
        let pathArray = path.split("/")
        let fileName = pathArray[pathArray.length - 1]
        let fileNameArray = fileName.split(".")
        let format = fileNameArray[fileNameArray.length - 1]
        blob.format = format
        stream.on("data", (data) => {
            hash.update(data)
        })

        stream.on("error", (err) => {
            stream.destroy()
            callback(new Errors.blobError("read uploadFile fail", { via: err }))
        })

        stream.on("end", () => {
            let hex = hash.digest("hex")
            //let hex = hash.update(Date.now() + blob.name).digest("hex")
            blob.hash = hex
            let newPath = `${this.path}/${hex}.${format}`
            fs.stat(newPath, (err, stats) => {
                if (err) {
                    fs.link(path, newPath, (err) => {
                        if (err) {
                            callback(new Errors.blobError("create link fail", { via: err }))
                            return
                        }
                        callback(null, blob)
                    })
                } else {
                    callback(null, blob)
                }
            })

        })
    }


    getBlob(option: { id: string }, callback: Callback<Blob>) {
        this.blobCollection.findOne({ id: option.id }, (err, blob) => {
            if (err || !blob) {
                callback(new Errors.notFound("not fond in mongodb"))
                return
            }
            callback(null, blob)
        })
    }

    getBlobDownloadUrl(option: { id: string }, callback) {
        this.getBlob({ id: option.id }, (err, result) => {
            if (err) {
                callback(err)
            } else {
                let route = "/" + this.route.split("/")[1] + `/${option.id}`
                let url = `${route}?download=true`
                callback(null, url)
            }
        })
    }

    getBlobUrl(blob: Blob, callback) {
        let blobId = blob.id
        this.blobCollection.findOne({ id: blobId }, (err, blob: Blob) => {
            if (err || !blob) {
                callback(new Errors.notFound("not font in mongodb"))
                return
            }
            callback(null, blob.url)
        })
    }
}
