import { Route, APIRoute, PageRoute, MiddlewareRoute, RouteService, Method } from "../../route"
import { ExpressService } from "../expressService"
import { LegacyUserService } from "../legacyUserService"
import { LegacyUserDefinitionService, LegacyUser } from "../legacyUserDefinitionService"
import { RouteSessionService } from "../routeSessionService"
import * as express from "express"
import * as Leaf from "leaf-ts"

declare const module
export const Errors = Leaf.ErrorDoc.create()
    .define("Forbidden")
    .define("AuthorizationFailed")
    .generate() as Leaf.ErrorsOf<{
        Forbidden: {}
        AuthorizationFailed: {}
    }>
export class LegacyUserAPIService<TLegacyUser extends LegacyUser = LegacyUser> extends RouteService {
    readonly name = "api.legacyUser"
    dependencies = ["ExpressService", "RouteSessionService", "LegacyUserService", "LegacyUserDefinitionService"]
    prefix = "/api/"
    services: {
        ExpressService: ExpressService
        RouteSessionService: RouteSessionService
        LegacyUserService: LegacyUserService<TLegacyUser>
        LegacyUserDefinitionService: LegacyUserDefinitionService<TLegacyUser>
    }
    initialize() {
        this.addRoutes(module.exports)
        this.installTo(this.services.ExpressService.server)
    }
}

export class SessionLegacyUserProvider extends MiddlewareRoute {
    services: {
        LegacyUserService: LegacyUserService<LegacyUser>
        RouteSessionService: RouteSessionService
    }
    req: express.Request & { session: { legacyUsername: string } }
    info: { legacyUser: LegacyUser }
    handle() {
        let session = this.req.session
        if (!session) {
            this.next()
        }
        let name = this.req.session.legacyUsername
        if (!name) {
            this.next()
            return
        }
        this.services.LegacyUserService.getLegacyUserByLegacyUsername(name, (err, legacyUser) => {
            if (err) {
                this.next()
                return
            }
            this.info.legacyUser = legacyUser
            this.next()
        })
    }
}

export class Register extends APIRoute<null> {
    readonly method = Method.POST
    readonly path = "legacyUser"
    readonly minLegacyUsernameLength = 3
    readonly maxLegacyUsernameLength = 32
    readonly minPasswordLength = 8
    readonly maxPasswordLength = 1024
    services: {
        LegacyUserService: LegacyUserService<LegacyUser>
        RouteSessionService: RouteSessionService
        LegacyUserDefinitionService: LegacyUserDefinitionService<LegacyUser>
    }
    body: Partial<LegacyUser>
    private checkLegacyUserValidation(legacyUser: Partial<LegacyUser>) {
        return this.services.LegacyUserDefinitionService.LegacyUserValidator.check(legacyUser)
    }
    req: express.Request & { session: { legacyUsername: string } }
    info: { legacyUser: LegacyUser }
    handle() {
        let legacyUserToCheck = this.body
        try {
            this.checkLegacyUserValidation(legacyUserToCheck)
        } catch (err) {
            this.error(err);
            return
        }
        this.services.LegacyUserService.createLegacyUser(legacyUserToCheck as LegacyUser, (err, legacyUser) => {
            if (err) {
                this.error(err)
                return
            }
            this.req.session.legacyUsername = legacyUser.legacyUsername
            this.success(null)
        })
    }
}

export class GetLegacyUserProfile extends APIRoute<LegacyUser>{
    readonly method = Method.GET
    readonly path = "legacyUser"
    services: {
        LegacyUserService: LegacyUserService<LegacyUser>
    }
    info: {
        legacyUser: LegacyUser
    }
    handle() {
        if (!this.info.legacyUser) {
            this.error(new Errors.Forbidden)
            return
        }
        let legacyUser = this.info.legacyUser
        this.success({
            legacyUsername: legacyUser.legacyUsername,
            email: legacyUser.email,
            displayName: legacyUser.displayName
        })
    }
}

export class Login extends APIRoute<LegacyUser>{
    readonly method = Method.PUT
    readonly path = "legacyUser"
    services: {
        LegacyUserService: LegacyUserService<LegacyUser>
    }
    body: {
        login: string
        password: string
    }
    req: express.Request & { session: { legacyUsername: string } }
    handle() {
        let loginLegacyUser = this.body
        this.services.LegacyUserService.getLegacyUserByLogin(loginLegacyUser, (err, legacyUser) => {
            if (err) {
                this.error(err)
                return
            }
            if (!legacyUser) {
                this.error(new Errors.AuthorizationFailed)
                return
            }
            this.req.session.legacyUsername = legacyUser.legacyUsername
            this.success(legacyUser)
        })
    }
}

export class Logout extends APIRoute<null> {
    readonly method = Method.DELETE
    readonly path = "legacyUser"
    req: express.Request & { session: { legacyUsername: string } }
    services: {
        LegacyUserService: LegacyUserService<LegacyUser>
    }
    handle() {
        this.req.session.legacyUsername = null
        this.success(null)
    }
}
